import numpy as np
import psyneulink as pnl
import pytest

from collections import OrderedDict

import psyneulink.core.components.functions.nonstateful.transferfunctions
from psyneulink.core.globals.keywords import ALLOCATION_SAMPLES, PROJECTIONS

class TestLog:

    def test_log(self):

        T_1 = pnl.TransferMechanism(name='log_test_T_1', input_shapes=2)
        T_2 = pnl.TransferMechanism(name='log_test_T_2', input_shapes=2)
        PS = pnl.Composition(name='log_test_PS', pathways=[T_1, T_2])
        PJ = T_2.path_afferents[0]

        assert T_1.loggable_items == {
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'InputPort-0': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'RESULT': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T_2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'execute_until_finished': 'OFF',
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_normalize': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_operation': 'OFF',
            'func_value': 'OFF',
            'has_initializers': 'OFF',
            'func_variable': 'OFF',
            'matrix': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_matrix': 'OFF',
            'num_executions_before_finished': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        T_1.set_log_conditions('mod_noise')
        T_1.set_log_conditions(pnl.RESULT)
        PJ.set_log_conditions('mod_matrix')

        assert T_1.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'EXECUTION',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'EXECUTION',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T_2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'execute_until_finished': 'OFF',
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_normalize': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_operation': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_matrix': 'EXECUTION',
            'num_executions_before_finished': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        PS.run(inputs={T_1:[0,0]})
        PS.run(inputs={T_1:[1,2]})
        PS.run(inputs={T_1:[3,4]})

        assert T_1.logged_items == {'RESULT': 'EXECUTION', 'mod_noise': 'EXECUTION'}
        assert PJ.logged_items == {'mod_matrix': 'EXECUTION'}

        T_1.log.print_entries(contexts=PS)
        # assert T_1.log.print_entries() ==
        # test_log.py::TestLog::test_log
        # Log for log_test_T_1:
        # Logged Item:   Time       Context                    Value
        # 'RESULT'       0:0:0:0   'PROCESSING, COMPOSI...   [0. 0.]
        # 'RESULT'       1:0:0:0   'PROCESSING, COMPOSI...   [1. 2.]
        # 'RESULT'       2:0:0:0   'PROCESSING, COMPOSI...   [3. 4.]
        # 'mod_noise'    0:0:0:0   'PROCESSING, COMPOSI...   [0.]
        # 'mod_noise'    1:0:0:0   'PROCESSING, COMPOSI...   [0.]
        # 'mod_noise'    2:0:0:0   'PROCESSING, COMPOSI...   [0.]

        T_1_csv = T_1.log.csv(entries=['mod_noise', 'RESULT'], owner_name=False, quotes=None)
        print(T_1_csv)
        assert T_1_csv == \
            "'Execution Context', 'Data'\n" \
            + f"'{PS.default_execution_id}', \'Run\', \'Trial\', \'Pass\', \'Time_step\', \'mod_noise\', \'RESULT\'\n" \
            + ", 0, 0, 0, 0, 0.0, 0.0 0.0\n" \
            + ", 1, 0, 0, 0, 0.0, 1.0 2.0\n" \
            + ", 2, 0, 0, 0, 0.0, 3.0 4.0\n"

        assert PJ.log.csv(entries='mod_matrix', owner_name=True, quotes=True) == \
            "'Execution Context', 'Data'\n" \
            + f"'{PS.default_execution_id}', \'Run\', \'Trial\', \'Pass\', \'Time_step\', " \
              "\'MappingProjection from log_test_T_1[RESULT] to log_test_T_2[InputPort-0][mod_matrix]\'\n" \
            + ", \'0\', \'0\', \'0\', \'1\', \'1.0 0.0\' \'0.0 1.0\'\n" \
            + ", \'1\', \'0\', \'0\', \'1\', \'1.0 0.0\' \'0.0 1.0\'\n" \
            + ", \'2\', \'0\', \'0\', \'1\', \'1.0 0.0\' \'0.0 1.0\'\n"

        result = T_1.log.nparray(entries=['mod_noise', 'RESULT'], header=False, owner_name=True)
        assert result[0] == PS.default_execution_id

        expected = [
            [[0], [1], [2]],
            [[0.], [0.], [0.]],
            [[0.], [0.], [0.]],
            [[0.], [0.], [0.]],
            [[0.0], [0.0], [0.0]],
            [[0., 0.], [1., 2.], [3., 4.]],
        ]
        assert result[1][0] == expected

    def test_log_initialization(self):
        T = pnl.TransferMechanism(
                prefs={pnl.LOG_PREF: pnl.PreferenceEntry(pnl.LogCondition.INITIALIZATION, pnl.PreferenceLevel.INSTANCE)}
        )
        assert T.logged_items == {'value': 'INITIALIZATION'}

    def test_log_dictionary_without_time(self):

        T1 = pnl.TransferMechanism(name='log_test_T1',
                                    input_shapes=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                    input_shapes=2)
        PS = pnl.Composition(name='log_test_PS', pathways=[T1, T2])
        PJ = T2.path_afferents[0]

        assert T1.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'num_executions_before_finished': 'OFF',
            'noise': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'execute_until_finished': 'OFF',
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_normalize': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_operation': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_matrix': 'OFF',
            'num_executions_before_finished': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULT)
        T1.set_log_conditions(pnl.VALUE)
        PJ.set_log_conditions('mod_matrix')
        T2.set_log_conditions('mod_slope')
        T2.set_log_conditions(pnl.RESULT)
        T2.set_log_conditions(pnl.VALUE)

        assert T1.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'EXECUTION',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'EXECUTION',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'execute_until_finished': 'OFF',
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_normalize': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_operation': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_matrix': 'EXECUTION',
            'num_executions_before_finished': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        PS.run(inputs={T1:[1.0, 2.0]})
        PS.run(inputs={T1:[3.0, 4.0]})
        PS.run(inputs={T1:[5.0, 6.0]})

        assert T1.logged_items == {'RESULT': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert T2.logged_items == {'RESULT': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert PJ.logged_items == {'mod_matrix': 'EXECUTION'}

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULT'])

        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert list(log_dict_T1.keys()) == [PS.default_execution_id]

        np.testing.assert_allclose(expected_values_T1, log_dict_T1[PS.default_execution_id]['value'])
        np.testing.assert_allclose(expected_slopes_T1, log_dict_T1[PS.default_execution_id]['mod_slope'])
        np.testing.assert_allclose(expected_results_T1, log_dict_T1[PS.default_execution_id]['RESULT'])

        assert list(log_dict_T1[PS.default_execution_id].keys()) == \
               ['Run', 'Trial', 'Pass', 'Time_step', 'value', 'mod_slope', 'RESULT']

        log_dict_T1_reorder = T1.log.nparray_dictionary(entries=['mod_slope', 'value', 'RESULT'])

        assert list(log_dict_T1_reorder[PS.default_execution_id].keys()) == \
               ['Run', 'Trial', 'Pass', 'Time_step', 'mod_slope', 'value', 'RESULT']

    def test_run_resets(self):
        import psyneulink as pnl
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   input_shapes=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   input_shapes=2)
        COMP = pnl.Composition(name='COMP', pathways=[T1, T2])
        T1.set_log_conditions('mod_slope')
        T2.set_log_conditions('value')
        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        log_array_T1 = T1.log.nparray()
        log_array_T2 = T2.log.nparray()
        assert log_array_T1.shape == log_array_T1.shape == (2,2)
        assert log_array_T1[0][0] == log_array_T2[0][0] == 'Execution Context'
        assert log_array_T1[0][1] == log_array_T1[0][1] == 'COMP'
        assert log_array_T2[1][1][4][1:4] == [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        log_array_T1_second_run = T1.log.nparray()
        log_array_T2_second_run = T2.log.nparray()
        assert log_array_T2_second_run[1][1][4][1:4] == [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]

    def test_log_dictionary_with_time(self):

        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   input_shapes=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.nonstateful.transferfunctions.Linear(slope=2.0),
                                   input_shapes=2)
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1, T2])

        assert T1.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'OFF',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }

        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULT)
        T1.set_log_conditions(pnl.VALUE)

        assert T1.loggable_items == {
            'execute_until_finished': 'OFF',
            'InputPort-0': 'OFF',
            'RESULT': 'EXECUTION',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }

        T2.set_log_conditions('mod_slope')
        T2.set_log_conditions(pnl.RESULT)
        T2.set_log_conditions(pnl.VALUE)

        assert T2.loggable_items == {
            'InputPort-0': 'OFF',
            'RESULT': 'EXECUTION',
            'clip': 'OFF',
            'termination_threshold': 'OFF',
            'execute_until_finished': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_execute_until_finished': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_max_executions_before_finished': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_num_executions_before_finished': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_executions_before_finished': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'mod_offset': 'OFF',
            'mod_rate': 'OFF',
            'noise': 'OFF',
            'num_executions_before_finished': 'OFF',
            'termination_measure_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }

        # RUN ZERO  |  TRIALS ZERO, ONE, TWO ----------------------------------

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        assert T1.logged_items == {'RESULT': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert T2.logged_items == {'RESULT': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}

        # T1 log after zero-th run -------------------------------------------

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULT'])

        expected_run_T1 = [[0], [0], [0]]
        expected_trial_T1 = [[0], [1], [2]]
        expected_time_step_T1 = [[0], [0], [0]]
        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert list(log_dict_T1.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T1[COMP.default_execution_id]

        np.testing.assert_allclose(expected_run_T1, sys_log_dict['Run'])
        np.testing.assert_allclose(expected_trial_T1, sys_log_dict['Trial'])
        np.testing.assert_allclose(expected_time_step_T1, sys_log_dict['Time_step'])
        np.testing.assert_allclose(expected_values_T1, sys_log_dict['value'])
        np.testing.assert_allclose(expected_slopes_T1, sys_log_dict['mod_slope'])
        np.testing.assert_allclose(expected_results_T1, sys_log_dict['RESULT'])

        # T2 log after zero-th run --------------------------------------------

        log_dict_T2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULT'])

        expected_run_T2 = [[0], [0], [0]]
        expected_trial_T2 = [[0], [1], [2]]
        expected_time_step_T2 = [[1], [1], [1]]
        expected_values_T2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]]
        expected_slopes_T2 = [[2.0], [2.0], [2.0]]
        expected_results_T2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]]

        assert list(log_dict_T2.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T2[COMP.default_execution_id]

        np.testing.assert_allclose(expected_run_T2, sys_log_dict['Run'])
        np.testing.assert_allclose(expected_trial_T2, sys_log_dict['Trial'])
        np.testing.assert_allclose(expected_time_step_T2, sys_log_dict['Time_step'])
        np.testing.assert_allclose(expected_values_T2, sys_log_dict['value'])
        np.testing.assert_allclose(expected_slopes_T2, sys_log_dict['mod_slope'])
        np.testing.assert_allclose(expected_results_T2, sys_log_dict['RESULT'])

        # RUN ONE  |  TRIALS ZERO, ONE, TWO -------------------------------------

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        # T1 log after first run -------------------------------------------

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULT'])

        assert list(log_dict_T1.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T1[COMP.default_execution_id]

        # expected_run_T1_2 = [[1], [1], [1]]
        expected_run_T1_2 = [[0], [0], [0]] + expected_run_T1
        expected_trial_T1_2 = [[0], [1], [2]] + expected_trial_T1
        expected_time_step_T1_2 = [[0], [0], [0]] + expected_time_step_T1
        expected_values_T1_2 = expected_values_T1 + expected_values_T1
        expected_slopes_T1_2 = expected_slopes_T1 + expected_slopes_T1
        expected_results_T1_2 = expected_results_T1 + expected_results_T1

        # np.testing.assert_allclose(expected_run_T1_2, sys_log_dict['Run'])
        # np.testing.assert_allclose(expected_trial_T1_2, sys_log_dict['Trial'])
        # np.testing.assert_allclose(expected_time_step_T1_2, sys_log_dict['Time_step'])
        np.testing.assert_allclose(expected_values_T1_2, sys_log_dict['value'])
        np.testing.assert_allclose(expected_slopes_T1_2, sys_log_dict['mod_slope'])
        np.testing.assert_allclose(expected_results_T1_2, sys_log_dict['RESULT'])

        # T2 log after first run -------------------------------------------

        log_dict_T2_2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULT'])

        assert list(log_dict_T2_2.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T2_2[COMP.default_execution_id]

        expected_run_T2_2 = [[0], [0], [0]] + expected_run_T2
        expected_trial_T2_2 = [[0], [1], [2]] + expected_trial_T2
        expected_time_step_T2_2 = [[1], [1], [1]] + expected_time_step_T2
        expected_values_T2_2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]] + expected_values_T2
        expected_slopes_T2_2 = [[2.0], [2.0], [2.0]] + expected_slopes_T2
        expected_results_T2_2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]] + expected_results_T2

        # np.testing.assert_allclose(expected_run_T2_2, sys_log_dict['Run'])
        # np.testing.assert_allclose(expected_trial_T2_2, sys_log_dict['Trial'])
        # np.testing.assert_allclose(expected_time_step_T2_2, sys_log_dict['Time_step'])
        np.testing.assert_allclose(expected_values_T2_2, sys_log_dict['value'])
        np.testing.assert_allclose(expected_slopes_T2_2, sys_log_dict['mod_slope'])
        np.testing.assert_allclose(expected_results_T2_2, sys_log_dict['RESULT'])

    def test_log_dictionary_with_scheduler(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.5)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.nonstateful.transferfunctions.Linear(slope=6.0))
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1, T2])

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.Threshold(T2, 'value', 5.0, '>=')
        }

        T1.set_log_conditions(pnl.VALUE)
        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULT)
        T2.set_log_conditions(pnl.VALUE)
        T2.set_log_conditions('mod_slope')

        COMP.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_dict_T1 = T1.log.nparray_dictionary(entries=['RESULT', 'mod_slope', 'value'])
        log_dict_T2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope'])

        assert list(log_dict_T1.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T1[COMP.default_execution_id]

        # Check order of keys (must match order of specification)
        assert list(sys_log_dict.keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'RESULT', 'mod_slope', 'value']
        assert list(log_dict_T2[COMP.default_execution_id].keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'value',
                                                                    'mod_slope']

        # Check values T1
        np.testing.assert_allclose(sys_log_dict["Run"], [[0], [0], [0]])
        np.testing.assert_allclose(sys_log_dict["Trial"], [[0], [0], [0]])
        np.testing.assert_allclose(sys_log_dict["Time_step"], [[0], [0], [0]])
        np.testing.assert_allclose(sys_log_dict["RESULT"], [[0.5], [0.75], [0.875]])
        np.testing.assert_allclose(sys_log_dict["value"], [[[0.5]], [[0.75]], [[0.875]]])
        np.testing.assert_allclose(sys_log_dict["mod_slope"], [[1], [1], [1]])

        # Check values T2
        np.testing.assert_allclose(log_dict_T2[COMP.default_execution_id]["Run"], [[0], [0], [0]])
        np.testing.assert_allclose(log_dict_T2[COMP.default_execution_id]["Trial"], [[0], [0], [0]])
        np.testing.assert_allclose(log_dict_T2[COMP.default_execution_id]["Time_step"], [[1], [1], [1]])
        np.testing.assert_allclose(log_dict_T2[COMP.default_execution_id]["value"], [[[3]], [[4.5]], [[5.25]]])
        np.testing.assert_allclose(log_dict_T2[COMP.default_execution_id]["mod_slope"], [[6], [6], [6]])

    def test_log_array_with_scheduler(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.5)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.nonstateful.transferfunctions.Linear(slope=6.0))
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1, T2])

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.Threshold(T2, 'value', 5.0, '>=')
        }

        T1.set_log_conditions(pnl.VALUE)
        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULT)
        T2.set_log_conditions(pnl.VALUE)
        T2.set_log_conditions('mod_slope')

        COMP.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_array_T1 = T1.log.nparray(entries=['RESULT', 'mod_slope', 'value'])
        log_array_T2 = T2.log.nparray(entries=['value', 'mod_slope'])

        context_results = [pnl.Log.context_header, COMP.default_execution_id]
        # Check values
        run_results = [["Run"], [0], [0], [0]]
        trial_results = [["Trial"], [0], [0], [0]]
        pass_results = [["Pass"], [0], [1], [2]]
        time_step_results = [["Time_step"], [0], [0], [0]]
        results_results = ["RESULT", [0.5], [0.75], [0.875]]
        slope_results = ["mod_slope", [1], [1], [1]]
        value_results = ["value", [[0.5]], [[0.75]], [[0.875]]]

        for i in range(2):
            assert log_array_T1[0][i] == context_results[i]

        assert log_array_T1[1][0] == pnl.Log.data_header
        data_array = log_array_T1[1][1]
        j = 0
        for i in range(4):
            assert data_array[j][i] == run_results[i]
            assert data_array[j + 1][i] == trial_results[i]
            assert data_array[j + 2][i] == pass_results[i]
            assert data_array[j + 3][i] == time_step_results[i]
            assert data_array[j + 4][i] == results_results[i]
            assert data_array[j + 5][i] == slope_results[i]
            assert data_array[j + 6][i] == value_results[i]

        # Check values
        run_results = [["Run"], [0], [0], [0]]
        trial_results = [["Trial"], [0], [0], [0]]
        pass_results = [["Pass"], [0], [1], [2]]
        time_step_results = [["Time_step"], [1], [1], [1]]
        value_results = ["value", [[3]], [[4.5]], [[5.25]]]
        slope_results = ["mod_slope", [6], [6], [6]]

        for i in range(2):
            assert log_array_T1[0][i] == context_results[i]

        assert log_array_T2[1][0] == pnl.Log.data_header
        data_array = log_array_T2[1][1]
        j = 0
        for i in range(4):
            assert data_array[j][i] == run_results[i]
            assert data_array[j + 1][i] == trial_results[i]
            assert data_array[j + 2][i] == pass_results[i]
            assert data_array[j + 3][i] == time_step_results[i]
            assert data_array[j + 4][i] == value_results[i]
            assert data_array[j + 5][i] == slope_results[i]

    def test_log_dictionary_with_scheduler_many_time_step_increments(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.05)
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1])

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.Threshold(T1, 'value', 0.95, '>=')
        }

        T1.set_log_conditions(pnl.VALUE)

        COMP.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value'])

        assert list(log_dict_T1.keys()) == [COMP.default_execution_id]
        sys_log_dict = log_dict_T1[COMP.default_execution_id]

        # Check order of keys (must match order of specification)
        assert list(sys_log_dict.keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'value']

        # # Check values T1
        assert len(sys_log_dict["Run"]) == 59
        np.testing.assert_allclose(sys_log_dict["Pass"][30], 30)
        np.testing.assert_allclose(sys_log_dict["Time_step"][30], 0)
        assert abs(sys_log_dict["value"][58]) >= 0.95
        assert abs(sys_log_dict["value"][57]) < 0.95

    def test_log_csv_multiple_contexts(self):
        A = pnl.TransferMechanism(name='A')
        B = pnl.TransferMechanism(name='B')
        C = pnl.TransferMechanism(name='C')

        C.set_log_conditions(pnl.VALUE)

        X = pnl.Composition(name='comp X')
        Y = pnl.Composition(name='comp Y')

        X.add_linear_processing_pathway([A, C])
        Y.add_linear_processing_pathway([B, C])

        # running with manual contexts for consistent output
        # because output is sorted by context
        X.run(inputs={A: 1}, context='comp X')
        Y.run(inputs={B: 2}, context='comp Y')

        expected_str = "'Execution Context', 'Data'\n" \
            + "'comp X', 'Run', 'Trial', 'Pass', 'Time_step', 'value'\n" \
            + ", '0', '0', '0', '1', '1.0'\n" \
            + "'comp Y', 'Run', 'Trial', 'Pass', 'Time_step', 'value'\n" \
            + ", '0', '0', '0', '1', '2.0'\n"
        assert C.log.csv() == expected_str

        print()
        print()

    @pytest.mark.parametrize(
            'scheduler_conditions, multi_run', [
                (False, False),
                (True, False),
                (True, True)
            ]
    )
    def test_log_multi_calls_single_timestep(self, scheduler_conditions, multi_run):
        lca = pnl.LCAMechanism(
                input_shapes=2,
                leak=0.5,
                threshold=0.515,
                reset_stateful_function_when=pnl.AtTrialStart()
        )
        lca.set_log_conditions(pnl.VALUE)
        m0 = pnl.ProcessingMechanism(
                input_shapes=2
        )
        comp = pnl.Composition()
        comp.add_linear_processing_pathway([m0, lca])
        if scheduler_conditions:
            comp.scheduler.add_condition(lca, pnl.AfterNCalls(m0, 2))
        comp.run(inputs={m0: [[1, 0], [1, 0], [1, 0]]})
        log_dict = lca.log.nparray_dictionary()['Composition-0']
        assert log_dict['Run'] == [[0], [0], [0]]
        assert log_dict['Trial'] == [[0], [1], [2]]
        assert log_dict['Pass'] == [[1], [1], [1]] if scheduler_conditions else [[0], [0], [0]]
        assert log_dict['Time_step'] == [[1], [1], [1]]
        # floats in value, so use np.allclose
        np.testing.assert_allclose(log_dict['value'], [[[0.52466739, 0.47533261]] for _ in range(3)])
        if multi_run:
            comp.run(inputs={m0: [[1, 0], [1, 0], [1, 0]]})
            log_dict = lca.log.nparray_dictionary()['Composition-0']
            assert log_dict['Run'] == [[0], [0], [0], [1], [1], [1]]
            np.testing.assert_allclose(log_dict['value'], [[[0.52466739, 0.47533261]] for _ in range(6)])

    def test_log_with_non_full_execution_id_entries(self):
        t = pnl.TransferMechanism()

        t.parameters.noise.set(0, context=1, override=True)
        t.parameters.value.set(0, context=2, override=True)

        t.log.nparray()
        t.log.nparray_dictionary()


class TestClearLog:

    def test_clear_log(self):

        # Create Composition
        T_1 = pnl.TransferMechanism(name='log_test_T_1', input_shapes=2)
        T_2 = pnl.TransferMechanism(name='log_test_T_2', input_shapes=2)
        COMP = pnl.Composition(name="log_test_COMP", pathways=[T_1, T_2])
        PJ = T_2.path_afferents[0]

        # Set log conditions on each component
        T_1.set_log_conditions('mod_noise')
        T_1.set_log_conditions(pnl.RESULT)
        T_2.set_log_conditions('mod_slope')
        T_2.set_log_conditions(pnl.RESULT)
        PJ.set_log_conditions('mod_matrix')

        # Run Composition
        COMP.run(inputs={T_1: [1.0, 1.0]})

        # Create log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        assert list(log_dict_T_1.keys()) == [COMP.default_execution_id]
        assert list(log_dict_T_2.keys()) == [COMP.default_execution_id]
        assert list(log_dict_PJ.keys()) == [COMP.default_execution_id]

        # Confirm that values were logged correctly
        sys_log_dict = log_dict_T_1[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['RESULT'], np.array([[1.0, 1.0]]))
        np.testing.assert_allclose(sys_log_dict['mod_noise'], np.array([[0.0]]))

        sys_log_dict = log_dict_T_2[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['RESULT'], np.array([[1.0, 1.0]]))
        np.testing.assert_allclose(sys_log_dict['mod_slope'], np.array([[1.0]]))

        sys_log_dict = log_dict_PJ[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['mod_matrix'], np.array([[[1.0, 0.0], [0.0, 1.0]]]))

        # KDM 10/3/18: below was changed to delete_entry=True because it's not implemented in Parameter logs,
        # and it's not clear this option results in much difference than just deleting the entries and
        # is stated to be included only for future use
        # Clear T_1s log and DO NOT delete entries
        T_1.log.clear_entries(delete_entry=True)

        # Clear T_2s log and delete entries
        T_2.log.clear_entries(delete_entry=True)

        # Create new log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        # Confirm that T_1 log values were removed
        assert log_dict_T_1 == OrderedDict()

        # Confirm that T_2 log values were removed and dictionary entries were destroyed
        assert log_dict_T_2 == OrderedDict()

        # Confirm that PJ log values were not affected by changes to T_1 and T_2's logs
        np.testing.assert_allclose(log_dict_PJ[COMP.default_execution_id]['mod_matrix'], np.array([[[1.0, 0.0], [0.0, 1.0]]]))

        # Run Composition again
        COMP.run(inputs={T_1: [2.0, 2.0]})

        # Create new log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        # Confirm that T_1 log values only include most recent run
        sys_log_dict = log_dict_T_1[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['RESULT'], np.array([[2.0, 2.0]]))
        np.testing.assert_allclose(sys_log_dict['mod_noise'], np.array([[0.0]]))
        # NOTE: "Run" value still incremented, but only the most recent one is returned (# runs does not reset to zero)
        np.testing.assert_allclose(sys_log_dict['Run'], np.array([[1]]))

        # Confirm that T_2 log values only include most recent run
        sys_log_dict = log_dict_T_2[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['RESULT'], np.array([[2.0, 2.0]]))
        np.testing.assert_allclose(sys_log_dict['mod_slope'], np.array([[1.0]]))
        np.testing.assert_allclose(sys_log_dict['Run'], np.array([[1]]))

        # Confirm that PJ log values include all runs
        sys_log_dict = log_dict_PJ[COMP.default_execution_id]
        np.testing.assert_allclose(sys_log_dict['mod_matrix'], np.array([[[1.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [0.0, 1.0]]]))
        np.testing.assert_allclose(sys_log_dict['Run'], np.array([[0], [1]]))

    @pytest.mark.parametrize(
        'insertion_eids, deletion_eids, log_is_empty',
        [
            (['context'], 'context', True),     # fails if string handling not correct due to str being Iterable
            (['context'], ['context'], True),
        ]
    )
    def test_clear_log_arguments(self, insertion_eids, deletion_eids, log_is_empty):
        t = pnl.TransferMechanism()
        c = pnl.Composition()
        c.add_node(t)

        t.parameters.value.log_condition = True

        for eid in insertion_eids:
            c.run({t: 0}, context=eid)

        t.parameters.value.clear_log(deletion_eids)

        if log_is_empty:
            assert len(t.parameters.value.log) == 0
        else:
            assert len(t.parameters.value.log) != 0


class TestFiltering:

    @pytest.fixture(scope='module')
    def node_logged_in_simulation(self):
        Input = pnl.TransferMechanism(name='Input')
        reward = pnl.TransferMechanism(
            output_ports=[pnl.RESULT, pnl.MEAN, pnl.VARIANCE], name='reward')
        Decision = pnl.DDM(
            function=pnl.DriftDiffusionAnalytical(
                drift_rate=(1.0, pnl.ControlProjection(
                    function=pnl.Linear,
                    control_signal_params={pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)})
                ),
                threshold=(1.0, pnl.ControlProjection(
                    function=pnl.Linear,
                    control_signal_params={pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)})
                ),
                noise=0.5,
                starting_value=0,
                non_decision_time=0.45
            ),
            output_ports=[
                pnl.DECISION_VARIABLE,
                pnl.RESPONSE_TIME,
                pnl.PROBABILITY_UPPER_THRESHOLD],
            name='Decision'
        )

        comp = pnl.Composition(name="evc", retain_old_simulation_data=True)
        comp.add_node(reward, required_roles=[pnl.NodeRole.OUTPUT])
        comp.add_node(Decision, required_roles=[pnl.NodeRole.OUTPUT])
        task_execution_pathway = [Input, pnl.IDENTITY_MATRIX, Decision]
        comp.add_linear_processing_pathway(task_execution_pathway)

        comp.add_controller(
            controller=pnl.OptimizationControlMechanism(
                agent_rep=comp,
                state_features=[Input.input_port, reward.input_port],
                state_feature_function=pnl.AdaptiveIntegrator(rate=0.5),
                objective_mechanism=pnl.ObjectiveMechanism(
                    function=pnl.LinearCombination(operation=pnl.PRODUCT),
                    monitor=[
                        reward,
                        Decision.output_ports[pnl.PROBABILITY_UPPER_THRESHOLD],
                        (Decision.output_ports[pnl.RESPONSE_TIME], -1, 1)
                    ]
                ),
                function=pnl.GridSearch(),
                control_signals=[
                    {PROJECTIONS: ("drift_rate", Decision),
                     ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)},
                    {PROJECTIONS: ("threshold", Decision),
                     ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)}
                ]
            )
        )

        comp.enable_controller = True

        stim_list_dict = {
            Input: [0.5, 0.123],
            reward: [20, 20]
        }

        Input.parameters.value.log_condition = True

        comp.run(inputs=stim_list_dict)

        return Input

    def test_node_has_logged_sims(self, node_logged_in_simulation):
        for logged_value, eid_dict in node_logged_in_simulation.log.logged_entries.items():
            for eid in eid_dict:
                if pnl.EID_SIMULATION in str(eid):
                    return
        else:
            assert False, 'No simulation execution_id found in log'

    def test_nparray(self, node_logged_in_simulation):
        for eid in node_logged_in_simulation.log.nparray(exclude_sims=True)[0]:
            assert pnl.EID_SIMULATION not in str(eid)

    def test_nparray_dictionary(self, node_logged_in_simulation):
        for eid in node_logged_in_simulation.log.nparray_dictionary(exclude_sims=True):
            assert pnl.EID_SIMULATION not in str(eid)

    def test_csv(self, node_logged_in_simulation):
        full_csv = node_logged_in_simulation.log.csv(exclude_sims=True)

        # get each row, excluding header
        for row in full_csv.split('\n')[1:]:
            # if present in a row, context will be in the first cell
            assert pnl.EID_SIMULATION not in row.replace("'", '').split(',')[0]


class TestFullModels:
    def test_multilayer(self):

        input_layer = pnl.TransferMechanism(name='input_layer',
                                            function=pnl.Logistic,
                                            input_shapes=2)

        hidden_layer_1 = pnl.TransferMechanism(name='hidden_layer_1',
                                               function=pnl.Logistic,
                                               input_shapes=5)

        hidden_layer_2 = pnl.TransferMechanism(name='hidden_layer_2',
                                               function=pnl.Logistic,
                                               input_shapes=4)

        output_layer = pnl.TransferMechanism(name='output_layer',
                                             function=pnl.Logistic,
                                             input_shapes=3)

        input_weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
        middle_weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
        output_weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)

        # This projection will be used by the process below by referencing it in the process' pathway;
        #    note: sender and receiver args don't need to be specified
        input_weights = pnl.MappingProjection(
            name='Input Weights',
            matrix=input_weights_matrix,
        )

        # This projection will be used by the process below by assigning its sender and receiver args
        #    to mechanismss in the pathway
        middle_weights = pnl.MappingProjection(
            name='Middle Weights',
            sender=hidden_layer_1,
            receiver=hidden_layer_2,
            matrix=middle_weights_matrix,
        )

        # Commented lines in this projection illustrate variety of ways in which matrix and learning signals can be specified
        output_weights = pnl.MappingProjection(
            name='Output Weights',
            sender=hidden_layer_2,
            receiver=output_layer,
            matrix=output_weights_matrix,
        )

        comp = pnl.Composition(name='multilayer')

        p = [input_layer, input_weights, hidden_layer_1, middle_weights, hidden_layer_2, output_weights, output_layer]
        backprop_pathway = comp.add_backpropagation_learning_pathway(
            pathway=p,
            loss_spec=pnl.Loss.L0,
            learning_rate=1.
        )

        input_dictionary = {backprop_pathway.target: [[0., 0., 1.]],
                            input_layer: [[-1., 30.]]}

        middle_weights.set_log_conditions(('mod_matrix', pnl.PROCESSING))

        comp.learn(inputs=input_dictionary,
                 num_trials=10)

        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                    [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]],
                    [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]],
                    [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]],
                    [[2], [2], [2], [2], [2], [2], [2], [2], [2], [2]],
                    [ [[ 0.05,  0.1 ,  0.15,  0.2 ],
                       [ 0.25,  0.3 ,  0.35,  0.4 ],
                       [ 0.45,  0.5 ,  0.55,  0.6 ],
                       [ 0.65,  0.7 ,  0.75,  0.8 ],
                       [ 0.85,  0.9 ,  0.95,  1.  ]],
                      [[ 0.04789907,  0.09413833,  0.14134241,  0.18938924],
                       [ 0.24780811,  0.29388455,  0.34096758,  0.38892985],
                       [ 0.44772121,  0.49364209,  0.54060947,  0.58849095],
                       [ 0.64763875,  0.69341202,  0.74026967,  0.78807449],
                       [ 0.84756101,  0.89319513,  0.93994932,  0.98768187]],
                      [[ 0.04738148,  0.08891106,  0.13248753,  0.177898  ],
                       [ 0.24726841,  0.28843403,  0.33173452,  0.37694783],
                       [ 0.44716034,  0.48797777,  0.53101423,  0.57603893],
                       [ 0.64705774,  0.6875443 ,  0.73032986,  0.77517531],
                       [ 0.84696096,  0.88713512,  0.92968378,  0.97435998]],
                      [[ 0.04937771,  0.08530344,  0.12439361,  0.16640433],
                       [ 0.24934878,  0.28467436,  0.32329947,  0.36496974],
                       [ 0.44932147,  0.48407216,  0.52225175,  0.56359587],
                       [ 0.64929589,  0.68349948,  0.72125508,  0.76228876],
                       [ 0.84927212,  0.88295836,  0.92031297,  0.96105307]],
                      [[ 0.05440291,  0.08430585,  0.1183739 ,  0.15641064],
                       [ 0.25458348,  0.28363519,  0.3170288 ,  0.35455942],
                       [ 0.45475764,  0.48299299,  0.51573974,  0.55278488],
                       [ 0.65492462,  0.68238209,  0.7145124 ,  0.75109483],
                       [ 0.85508376,  0.88180465,  0.91335119,  0.94949538]],
                      [[ 0.06177218,  0.0860581 ,  0.11525064,  0.14926369],
                       [ 0.26225812,  0.28546004,  0.31377611,  0.34711631],
                       [ 0.46272625,  0.48488774,  0.51236246,  0.54505667],
                       [ 0.66317453,  0.68434373,  0.7110159 ,  0.74309381],
                       [ 0.86360121,  0.88382991,  0.9097413 ,  0.94123489]],
                      [[ 0.06989398,  0.08959148,  0.11465594,  0.14513241],
                       [ 0.27071639,  0.2891398 ,  0.31315677,  0.34281389],
                       [ 0.47150846,  0.48870843,  0.5117194 ,  0.54058946],
                       [ 0.67226675,  0.68829929,  0.71035014,  0.73846891],
                       [ 0.87298831,  0.88791376,  0.90905395,  0.93646   ]],
                      [[ 0.07750784,  0.09371987,  0.11555569,  0.143181  ],
                       [ 0.27864693,  0.29343991,  0.31409396,  0.3407813 ],
                       [ 0.47974374,  0.49317377,  0.5126926 ,  0.53847878],
                       [ 0.68079346,  0.69292265,  0.71135777,  0.73628353],
                       [ 0.88179203,  0.89268732,  0.91009431,  0.93420362]],
                      [[ 0.0841765 ,  0.09776672,  0.11711835,  0.14249779],
                       [ 0.28559463,  0.29765609,  0.31572199,  0.34006951],
                       [ 0.48695967,  0.49755273,  0.51438349,  0.5377395 ],
                       [ 0.68826567,  0.69745713,  0.71310872,  0.735518  ],
                       [ 0.88950757,  0.89736946,  0.91190228,  0.93341316]],
                      [[ 0.08992499,  0.10150104,  0.11891032,  0.14250149],
                       [ 0.29158517,  0.30154765,  0.31758943,  0.34007336],
                       [ 0.49318268,  0.50159531,  0.51632339,  0.5377435 ],
                       [ 0.69471052,  0.70164382,  0.71511777,  0.73552215],
                       [ 0.8961628 ,  0.90169281,  0.91397691,  0.93341744]]]
                ]]
            ],
            dtype=object
        )
        log_val = middle_weights.log.nparray(entries='mod_matrix', header=False)

        assert log_val[0] == expected_log_val[0]

        for i in range(1, len(log_val)):
            try:
                np.testing.assert_allclose(log_val[i], expected_log_val[i])
            except TypeError:
                for j in range(len(log_val[i])):
                    np.testing.assert_allclose(
                        np.array(log_val[i][j][0]),
                        np.array(expected_log_val[i][j][0]),
                        atol=1e-08,
                        err_msg='Failed on test item {0} of logged values'.format(i)
                    )

        middle_weights.log.print_entries()

        # Test Programmatic logging
        hidden_layer_2.log.log_values(pnl.VALUE, comp)
        log_val = hidden_layer_2.log.nparray(header=False)
        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                    [[1]],
                    [[0]],
                    [[0]],
                    [[0]],
                    [[[0.8565238418942037, 0.8601053239957609, 0.8662098921116546, 0.8746933736954071]]]
                ]]
            ],
            dtype=object
        )
        assert log_val[0] == expected_log_val[0]

        for i in range(1, len(log_val)):
            try:
                np.testing.assert_allclose(log_val[i], expected_log_val[i])
            except TypeError:
                for j in range(len(log_val[i])):
                    np.testing.assert_allclose(
                        np.array(log_val[i][j][0]),
                        np.array(expected_log_val[i][j][0]),
                        atol=1e-08,
                        err_msg='Failed on test item {0} of logged values'.format(i)
                    )
        hidden_layer_2.log.print_entries()

        # Clear log and test with logging of weights set to LEARNING for another 5 trials of learning
        middle_weights.log.clear_entries(entries=None, confirm=False)
        middle_weights.set_log_conditions(('mod_matrix', pnl.LEARNING))
        comp.learn(
            num_trials=5,
            inputs=input_dictionary,
        )
        log_val = middle_weights.log.nparray(entries='mod_matrix', header=False)
        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                        [[1], [1], [1], [1], [1]],  # RUN
                        # [[0], [1], [2], [3], [4]],  # TRIAL
                        [[0], [1], [2], [3], [4]],  # TRIAL
                        [[1], [1], [1], [1], [1]],  # PASS
                        # [[0], [0], [0], [0], [0]],  # PASS
                        # [[1], [1], [1], [1], [1]],  # TIME_STEP
                        [[0], [0], [0], [0], [0]],  # TIME_STEP
                        [  [[0.09925812411381937, 0.1079522130303428, 0.12252820028789306, 0.14345816973727732],
                            [0.30131473371328343, 0.30827285172236585, 0.3213609999139731, 0.3410707131678078],
                            [0.5032924245149345, 0.5085833053183328, 0.5202423523987703, 0.5387798509126243],
                            [0.70518251216691, 0.7088822116145151, 0.7191771716324874, 0.7365956448426355],
                            [0.9069777724600303, 0.9091682860319945, 0.9181692763668221, 0.93452610920817]],
                           [[0.103113468050986, 0.11073719161508278, 0.12424368674464399, 0.14415219181047598],
                            [0.3053351724284921, 0.3111770895557729, 0.3231499474835138, 0.341794454877438],
                            [0.5074709829757806, 0.5116017638574931, 0.5221016574478528, 0.5395320566440044],
                            [0.7095115080472698, 0.7120093413898914, 0.7211034158081356, 0.7373749316571768],
                            [0.9114489813353512, 0.9123981459792809, 0.9201588001021687, 0.935330996581107]],
                          [[0.10656261740658036, 0.11328192907953168, 0.12587702586370172, 0.14490737831188183],
                           [0.30893272045369513, 0.31383131362555394, 0.32485356055342113, 0.3425821330631872],
                           [0.5112105492674988, 0.5143607671543178, 0.5238725230390068, 0.5403508295336265],
                           [0.7133860755337162, 0.7148679468096026, 0.7229382109974996, 0.7382232628724675],
                           [0.9154510531345043, 0.9153508224199809, 0.9220539747533424, 0.936207244690072]],
                          [[0.10967776822419642, 0.11562091141141007, 0.12742795007904037, 0.14569308665620523],
                           [0.3121824816018084, 0.316271366885665, 0.3264715025259811, 0.34340179304134666],
                           [0.5145890402653069, 0.5168974760377518, 0.5255545550838675, 0.5412029579613059],
                           [0.7168868378231593, 0.7174964619674593, 0.7246811176253708, 0.7391062307617761],
                           [0.9190671994078436, 0.9180659725806082, 0.923854327015523, 0.9371193149131859]],
                          [[0.11251466428344682, 0.11778293740676549, 0.12890014813698167, 0.14649079441816393],
                           [0.31514245505635713, 0.3185271913574249, 0.328007571201157, 0.3442341089776976],
                           [0.5176666356203712, 0.5192429413004418, 0.5271516632648602, 0.5420683480396268],
                           [0.7200760707077265, 0.7199270072739019, 0.7263361597421493, 0.7400030122347587],
                           [0.922361699102421, 0.9205767427437028, 0.9255639970037588, 0.9380456963960624]]]
                ]]
            ],
            dtype=object
        )

        assert log_val.shape == expected_log_val.shape
        assert log_val[0] == expected_log_val[0]
        assert len(log_val[1]) == len(expected_log_val[1]) == 1

        for i in range(len(log_val[1][0])):
            try:
                np.testing.assert_allclose(
                    log_val[1][0][i],
                    expected_log_val[1][0][i],
                    err_msg='Failed on test item {0} of logged values'.format(i)
                )
            except TypeError:
                for j in range(len(log_val[1][0][i])):
                    np.testing.assert_allclose(
                        np.array(log_val[1][0][i][j]),
                        np.array(expected_log_val[1][0][i][j]),
                        atol=1e-08,
                        err_msg='Failed on test item {0} of logged values'.format(i)
                    )



"""
This implements a model of mesolimbic dopamine cell activity during monkey
conditioning as found in `Montague, Dayan, and Sejnowski (1996) in PsyNeuLink
<http://www.jneurosci.org/content/jneuro/16/5/1936.full.pdf>`_
"""
import argparse

import numpy as np
import psyneulink as pnl

all_figures = ['5a', '5b', '5c']

parser = argparse.ArgumentParser()
parser.add_argument('--no-plot', action='store_false', help='Disable plotting', dest='enable_plot')
parser.add_argument('--figure', nargs='+', help='Figure(s) to plot (default=all)', choices=all_figures, default=all_figures)
args = parser.parse_args()

if args.enable_plot:
    import matplotlib.pyplot as plt


def build_stimulus_dictionary(sample_mechanism, target_mechanism, no_reward_trials):

    stimulus_onset = 41
    reward_delivery = 54

    samples = []
    targets = []
    for trial in range(120):
        target = [0.] * 60
        target[reward_delivery] = 1.
        if trial in no_reward_trials:
            target[reward_delivery] = 0.
        targets.append(target)

        sample = [0.] * 60
        for i in range(stimulus_onset, 60):
            sample[i] = 1.
        samples.append(sample)

    return {sample_mechanism: samples,
            target_mechanism: targets}


def build_stimulus_dictionary_figure_5c(sample_mechanism, target_mechanism):

    stimulus_onset = 42
    reward_delivery = 54

    # build input dictionary
    samples = []
    targets = []
    for trial in range(150):
        target = [0.] * 60
        target[reward_delivery] = 1.
        if trial > 70:
            target[reward_delivery] = 0.
        targets.append(target)

        sample = [0.] * 60
        for i in range(stimulus_onset, 60):
            sample[i] = 1.
        samples.append(sample)

    return {sample_mechanism: samples,
            target_mechanism: targets}


def figure_5a():
    """
    This creates the plot for figure 5A in the Montague paper. Figure 5A is
    a 'plot of ∂(t) over time for three trials during training (1, 30, and 50).'
    """

    # Create Processing Components
    sample_mechanism = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             name=pnl.SAMPLE)

    action_selection = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             function=pnl.Linear(slope=1.0,
                                                                 intercept=0.01),
                                             name='Action Selection')

    sample_to_action_selection = pnl.MappingProjection(sender=sample_mechanism,
                                                       receiver=action_selection,
                                                       matrix=np.zeros((60, 60)))
    # Create Composition
    composition_name = 'TD_Learning_Figure_5A'
    comp = pnl.Composition(name=composition_name)

    # Add Processing Components to the Composition
    pathway = [sample_mechanism, sample_to_action_selection, action_selection]

    # Add Learning Components to the Composition
    learning_related_components = comp.add_td_learning_pathway(pathway, learning_rate=0.3).learning_components

    # Unpack Relevant Learning Components
    prediction_error_mechanism = learning_related_components[pnl.OBJECTIVE_MECHANISM]
    target_mechanism = learning_related_components[pnl.TARGET_MECHANISM]

    # Create Log
    prediction_error_mechanism.log.set_log_conditions(pnl.VALUE)

    # Create Stimulus Dictionary
    no_reward_trials = {14, 29, 44, 59, 74, 89}
    inputs = build_stimulus_dictionary(sample_mechanism, target_mechanism, no_reward_trials)

    # Run Composition
    comp.learn(inputs=inputs)

    if args.enable_plot:
        # Get Delta Values from Log
        delta_vals = prediction_error_mechanism.log.nparray_dictionary()[composition_name][pnl.VALUE]

        # Plot Delta Values form trials 1, 30, and 50
        with plt.style.context('seaborn-v0_8'):
            plt.plot(delta_vals[0][0], "-o", label="Trial 1")
            plt.plot(delta_vals[29][0], "-s", label="Trial 30")
            plt.plot(delta_vals[49][0], "-o", label="Trial 50")
            plt.title("Montague et. al. (1996) -- Figure 5A")
            plt.xlabel("Timestep")
            plt.ylabel("∂")
            plt.legend()
            plt.xlim(xmin=35)
            plt.xticks()
            plt.show(block=not pnl._called_from_pytest)

    return comp


def figure_5b():
    """
    This creates the plot for figure 5B in the Montague paper. Figure 5B shows
    the 'entire time course of model responses (trials 1-150).' The setup is
    the same as in Figure 5A, except that training begins at trial 10.
    """

    # Create Processing Components
    sample_mechanism = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             name=pnl.SAMPLE)

    action_selection = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             function=pnl.Linear(slope=1.0,
                                                                 intercept=1.0),
                                             name='Action Selection')

    sample_to_action_selection = pnl.MappingProjection(sender=sample_mechanism,
                                                       receiver=action_selection,
                                                       matrix=np.zeros((60, 60)))
    # Create Composition
    composition_name = 'TD_Learning_Figure_5B'
    comp = pnl.Composition(name=composition_name)

    # Add Processing Components to the Composition
    pathway = [sample_mechanism, sample_to_action_selection, action_selection]

    # Add Learning Components to the Composition
    learning_related_components = comp.add_td_learning_pathway(pathway, learning_rate=0.3).learning_components

    # Unpack Relevant Learning Components
    prediction_error_mechanism = learning_related_components[pnl.OBJECTIVE_MECHANISM]
    target_mechanism = learning_related_components[pnl.TARGET_MECHANISM]

    # Create Log
    prediction_error_mechanism.log.set_log_conditions(pnl.VALUE)

    # Create Stimulus Dictionary
    no_reward_trials = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 14, 29, 44, 59, 74,
                        89, 104, 119}
    inputs = build_stimulus_dictionary(sample_mechanism, target_mechanism, no_reward_trials)

    # Run Composition
    comp.learn(inputs=inputs)

    if args.enable_plot:
        # Get Delta Values from Log
        delta_vals = prediction_error_mechanism.log.nparray_dictionary()[composition_name][pnl.VALUE]

        with plt.style.context('seaborn-v0_8'):
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            x_vals, y_vals = np.meshgrid(np.arange(120), np.arange(40, 60, step=1))
            d_vals = np.array([d[0][40:60] for d in delta_vals]).transpose()
            ax.plot_surface(x_vals, y_vals, d_vals)
            ax.set_xlabel("Trial")
            ax.set_ylabel("Timestep")
            ax.set_zlabel("∂")
            ax.set_ylim(y_vals.max(), y_vals.min())
            ax.set_xlim(0, 120)
            ax.set_zlim(-1, 1)
            ax.set_title("Montague et. al. (1996) -- Figure 5B")
            plt.show(block=not pnl._called_from_pytest)

    return comp


def figure_5c():
    """
    This creates the plot for Figure 5C in the Montague paper. Figure 5C shows
    'extinction of response to the sensory cue.' The setup is the same as
    Figure 5A, except that reward delivery stops at trial 70
    """

    # Create Processing Components
    sample_mechanism = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             name=pnl.SAMPLE)

    action_selection = pnl.TransferMechanism(default_variable=np.zeros(60),
                                             function=pnl.Linear(slope=1.0,
                                                                 intercept=1.0),
                                             name='Action Selection')

    sample_to_action_selection = pnl.MappingProjection(sender=sample_mechanism,
                                                       receiver=action_selection,
                                                       matrix=np.zeros((60, 60)))
    # Create Composition
    composition_name = 'TD_Learning_Figure_5C'
    comp = pnl.Composition(name=composition_name)

    # Add Processing Components to the Composition
    pathway = [sample_mechanism, sample_to_action_selection, action_selection]

    # Add Learning Components to the Composition
    learning_related_components = comp.add_td_learning_pathway(pathway, learning_rate=0.3).learning_components

    # Unpack Relevant Learning Components
    prediction_error_mechanism = learning_related_components[pnl.OBJECTIVE_MECHANISM]
    target_mechanism = learning_related_components[pnl.TARGET_MECHANISM]

    # Create Log
    prediction_error_mechanism.log.set_log_conditions(pnl.VALUE)

    # Create Stimulus Dictionary
    inputs = build_stimulus_dictionary_figure_5c(sample_mechanism, target_mechanism)

    # Run Composition
    comp.learn(inputs=inputs)

    if args.enable_plot:
        # Get Delta Values from Log
        delta_vals = prediction_error_mechanism.log.nparray_dictionary()[composition_name][pnl.VALUE]

        with plt.style.context('seaborn-v0_8'):
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            x_vals, y_vals = np.meshgrid(np.arange(150), np.arange(40, 60, step=1))
            d_vals = np.array([d[0][40:60] for d in delta_vals]).transpose()
            ax.plot_surface(x_vals, y_vals, d_vals)
            ax.set_ylim(y_vals.max(), y_vals.min())
            ax.set_xlim(0, 140)
            ax.set_zlim(-1, 1)
            ax.set_xlabel("Trial")
            ax.set_ylabel("Timestep")
            ax.set_zlabel("∂")
            ax.set_title("Montague et. al. (1996) -- Figure 5C")
            plt.show(block=not pnl._called_from_pytest)

    return comp


if '5a' in args.figure:
    comp_5a = figure_5a()

if '5b' in args.figure:
    comp_5b = figure_5b()

if '5c' in args.figure:
    comp_5c = figure_5c()



import numpy as np
import psyneulink as pnl


# function that computes the accuracy of the model based on the inputLayer (task dimension) and stimulusInfo (stimulus
# of each task dimension). EX.If the task dimension is "Color", and the stimulus in the color dimension is 1 (positive),
# the associated accuracy of the trial will be the probability that the DDM hits the upper threshold
def computeAccuracy(variable):

    # variable is the list of values given by the monitored output ports in the Objective Mechanism

    print("Inputs to ComputeAccuracy Function: ", variable)

    taskInputs = variable[0]
    stimulusInputs = variable[1]
    upperThreshold = variable[2]
    lowerThreshold = variable[3]

    accuracy = []
    colorTrial = (taskInputs[0] == 1)
    motionTrial = (taskInputs[1] == 1)

    # during color trials

    if colorTrial:
        # if the correct answer is the upper threshold
        if stimulusInputs[0] == 1:
            accuracy.append(upperThreshold)
            print('Color Trial: 1')

        # if the correct answer is the lower threshold
        elif stimulusInputs[0] == -1:
            accuracy.append(lowerThreshold)
            print('Color Trial: -1')

    if motionTrial:
        # if the correct answer is the upper threshold
        if stimulusInputs[1] == 1:
            accuracy.append(upperThreshold)
            print('Motion Trial: 1')

        # if the correct answer is the lower threshold
        elif stimulusInputs[1] == -1:
            accuracy.append(lowerThreshold)
            print('Motion Trial: -1')

    # added in after original "concept" for function to account for simulations where the variable does pass in inputs
    # as expected, which is exactly the problem we're currently trying to solve
    if len(accuracy) == 0:
        accuracy = [0]
        # print('No Input')

    return accuracy



##### BEGIN STABILITY FLEXIBILITY MODEL CONSTRUCTION

tau = 0.9 # time constant
g = 1

# testing input
INPUT = [[1, 0], [1, 0], [1, 0], [0, 1], [0, 1], [0, 1]]
stimulusInput = [[1, -1], [1, -1], [1, -1], [1, -1], [1, -1], [1, -1]]
runs = len(INPUT)

excitatoryWeight = np.asarray([[1]])
inhibitoryWeight = np.asarray([[-1]])
gain = np.asarray([[g]])

DRIFT = 1 # Drift Rate
STARTING_VALUE = 0.0 # Starting Point
THRESHOLD = 0.0475 # Threshold
NOISE = 0.04 # Noise
T0 = 0.2 # T0


# first element is color task attendance, second element is motion task attendance
inputLayer = pnl.TransferMechanism(#default_variable=[[0.0, 0.0]],
                                   input_shapes=2,
                                   function=pnl.Linear(slope=1, intercept=0),
                                   output_ports = [pnl.RESULT],
                                   name='Input')
inputLayer.set_log_conditions([pnl.RESULT])

# Recurrent Transfer Mechanism that models the recurrence in the activation between the two stimulus and action
# dimensions. Positive self excitation and negative opposite inhibition with an integrator rate = tau
# Modulated variable in simulations is the GAIN variable of this mechanism
activation = pnl.RecurrentTransferMechanism(default_variable=[[0.0, 0.0]],
                                            function=pnl.Logistic(gain=1.0),
                                            matrix=[[1.0, -1.0],
                                                    [-1.0, 1.0]],
                                            integrator_mode = True,
                                            integrator_function=pnl.AdaptiveIntegrator(rate=(tau)),
                                            initial_value=np.array([[0.0, 0.0]]),
                                            output_ports = [pnl.RESULT],
                                            name = 'Activity')

activation.set_log_conditions([pnl.RESULT, "mod_gain"])


stimulusInfo = pnl.TransferMechanism(default_variable=[[0.0, 0.0]],
                                     input_shapes= 2,
                                     function = pnl.Linear(slope=1, intercept=0),
                                     output_ports = [pnl.RESULT],
                                     name = "Stimulus Info")

stimulusInfo.set_log_conditions([pnl.RESULT])

controlledElement = pnl.TransferMechanism(default_variable=[[0.0, 0.0]],
                                          input_shapes= 2,
                                          function=pnl.Linear(slope=1, intercept= 0),
                                          input_ports=pnl.InputPort(combine=pnl.PRODUCT),
                                          output_ports = [pnl.RESULT],
                                          name = 'Stimulus Info * Activity')

controlledElement.set_log_conditions([pnl.RESULT])

ddmCombination = pnl.TransferMechanism(
    input_shapes= 1,
                                       function = pnl.Linear(slope=1, intercept=0),
                                       output_ports = [pnl.RESULT],
                                       name = "DDM Integrator")
ddmCombination.set_log_conditions([pnl.RESULT])

decisionMaker = pnl.DDM(function=pnl.DriftDiffusionAnalytical(drift_rate = DRIFT,
                                                                 starting_value = STARTING_VALUE,
                                                                 threshold = THRESHOLD,
                                                                 noise = NOISE,
                                                                 non_decision_time = T0),
                                                                 output_ports = [pnl.DECISION_VARIABLE, pnl.RESPONSE_TIME,
                                                                                  pnl.PROBABILITY_UPPER_THRESHOLD, pnl.PROBABILITY_LOWER_THRESHOLD],
                                                                 name='DDM')

decisionMaker.set_log_conditions([pnl.PROBABILITY_UPPER_THRESHOLD, pnl.PROBABILITY_LOWER_THRESHOLD,
                            pnl.DECISION_VARIABLE, pnl.RESPONSE_TIME])

########### Composition

stabilityFlexibility = pnl.Composition()

### NODE CREATION

stabilityFlexibility.add_node(inputLayer)
stabilityFlexibility.add_node(activation)
stabilityFlexibility.add_node(controlledElement)
stabilityFlexibility.add_node(stimulusInfo)
stabilityFlexibility.add_node(ddmCombination)
stabilityFlexibility.add_node(decisionMaker)


stabilityFlexibility.add_projection(sender= inputLayer, receiver = activation)
stabilityFlexibility.add_projection(sender = activation, receiver = controlledElement)
stabilityFlexibility.add_projection(sender = stimulusInfo, receiver = controlledElement)
stabilityFlexibility.add_projection(sender = stimulusInfo, receiver = ddmCombination)
stabilityFlexibility.add_projection(sender = controlledElement, receiver = ddmCombination)
stabilityFlexibility.add_projection(sender = ddmCombination, receiver = decisionMaker)

# beginning of Controller

search_range = pnl.SampleSpec(start=1.0, stop=1.3, num=3)

signal = pnl.ControlSignal(modulates=[(pnl.GAIN, activation)],
                           function=pnl.Linear,
                           variable=1.0,
                           allocation_samples=search_range)

objective_mech = pnl.ObjectiveMechanism(monitor = [inputLayer, stimulusInfo,
                                                           (pnl.PROBABILITY_UPPER_THRESHOLD, decisionMaker),
                                                           (pnl.PROBABILITY_LOWER_THRESHOLD, decisionMaker)],
                                                           function = computeAccuracy)

meta_controller = pnl.OptimizationControlMechanism(agent_rep = stabilityFlexibility,
                                                   state_features= [inputLayer.input_port, stimulusInfo.input_port],
                                                   # state_features = {pnl.SHADOW_INPUTS: [inputLayer, stimulusInfo]},
                                                   # state_features = [(inputLayer, pnl.SHADOW_INPUTS),
                                                   #             (stimulusInfo, pnl.SHADOW_INPUTS)],
                                                   objective_mechanism = objective_mech,
                                                   function = pnl.GridSearch(),
                                                   control_signals = [signal])

inputs = {inputLayer: INPUT, stimulusInfo: stimulusInput}
stabilityFlexibility.add_model_based_optimizer(meta_controller)
stabilityFlexibility.enable_model_based_optimizer = True

# stabilityFlexibility.show_graph()

print("Beginning of Run")
stabilityFlexibility.run(inputs)
print(stabilityFlexibility.results)



import ctypes
import ctypes.util
import numpy as np
import pytest
import sys

from psyneulink.core import llvm as pnlvm
from llvmlite import ir


DIM_X = 1000
TST_MIN = 1.0
TST_MAX = 3.0

VECTOR = np.random.rand(DIM_X)

@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
def test_helper_fclamp(mode):

    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        double_ptr_ty = ir.DoubleType().as_pointer()
        func_ty = ir.FunctionType(ir.VoidType(), (double_ptr_ty, ctx.int32_ty,
                                                  double_ptr_ty))

        # Create clamp function
        custom_name = ctx.get_unique_name("clamp")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        vec, count, bounds = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        tst_min = builder.load(builder.gep(bounds, [ctx.int32_ty(0)]))
        tst_max = builder.load(builder.gep(bounds, [ctx.int32_ty(1)]))

        index = None
        with pnlvm.helpers.for_loop_zero_inc(builder, count, "linear") as (b1, index):
            val_ptr = b1.gep(vec, [index])
            val = b1.load(val_ptr)
            val = pnlvm.helpers.fclamp(b1, val, tst_min, tst_max)
            b1.store(val, val_ptr)

        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name, ctype_ptr_args=(0, 2))
    local_vec = VECTOR.copy()

    ref = np.clip(VECTOR, TST_MIN, TST_MAX)
    bounds = np.asarray([TST_MIN, TST_MAX], dtype=bin_f.np_arg_dtypes[2].base)

    if mode == 'CPU':
        ct_vec = local_vec.ctypes.data_as(bin_f.c_func.argtypes[0])
        ct_bounds = bounds.ctypes.data_as(bin_f.c_func.argtypes[2])

        bin_f(ct_vec, DIM_X, ct_bounds)
    else:
        bin_f.cuda_wrap_call(local_vec, np.int32(DIM_X), bounds)

    assert np.array_equal(local_vec, ref)


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
def test_helper_fclamp_const(mode):

    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        double_ptr_ty = ir.DoubleType().as_pointer()
        func_ty = ir.FunctionType(ir.VoidType(), (double_ptr_ty, ctx.int32_ty))

        # Create clamp function
        custom_name = ctx.get_unique_name("clamp")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        vec, count = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        index = None
        with pnlvm.helpers.for_loop_zero_inc(builder, count, "linear") as (b1, index):
            val_ptr = b1.gep(vec, [index])
            val = b1.load(val_ptr)
            val = pnlvm.helpers.fclamp(b1, val, TST_MIN, TST_MAX)
            b1.store(val, val_ptr)

        builder.ret_void()

    local_vec = VECTOR.copy()
    ref = np.clip(VECTOR, TST_MIN, TST_MAX)

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name, ctype_ptr_args=(0,))
    if mode == 'CPU':
        ct_vec = local_vec.ctypes.data_as(bin_f.c_func.argtypes[0])

        bin_f(ct_vec, DIM_X)
    else:
        bin_f.cuda_wrap_call(local_vec, np.int32(DIM_X))

    assert np.array_equal(local_vec, ref)


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('rtol,atol', [[0, 0], [None, None], [None, 100], [2, None]])
@pytest.mark.parametrize('var1,var2',
                         [[1, 1], [1, 100], [1,2], [-4,5], [0, -100], [-1,-2],
                          [[1,1,1,-4,0,-1], [1,100,2,5,-100,-2]]
                         ])
@pytest.mark.parametrize('fp_type', [ir.DoubleType, ir.FloatType])
def test_helper_is_close(mode, var1, var2, rtol, atol, fp_type):

    # Instantiate LLVMBuilderContext using the preferred fp type
    pnlvm.builder_context.LLVMBuilderContext(fp_type())

    tolerance = {}
    if rtol is not None:
        tolerance['rtol'] = rtol
    if atol is not None:
        tolerance['atol'] = atol

    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        float_ptr_ty = ctx.float_ty.as_pointer()
        func_ty = ir.FunctionType(ir.VoidType(), [float_ptr_ty, float_ptr_ty, float_ptr_ty, ctx.int32_ty])

        custom_name = ctx.get_unique_name("is_close")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        in1, in2, out, count = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        index = None
        with pnlvm.helpers.for_loop_zero_inc(builder, count, "compare") as (b1, index):
            val1_ptr = b1.gep(in1, [index])
            val2_ptr = b1.gep(in2, [index])
            val1 = b1.load(val1_ptr)
            val2 = b1.load(val2_ptr)
            close = pnlvm.helpers.is_close(ctx, b1, val1, val2, **tolerance)
            out_ptr = b1.gep(out, [index])
            out_val = b1.select(close, out_ptr.type.pointee(1), out_ptr.type.pointee(0))
            b1.store(out_val, out_ptr)

        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name, ctype_ptr_args=(0, 1, 2))

    vec1 = np.atleast_1d(np.asarray(var1, dtype=bin_f.np_arg_dtypes[0].base))
    vec2 = np.atleast_1d(np.asarray(var2, dtype=bin_f.np_arg_dtypes[1].base))
    assert len(vec1) == len(vec2)
    res = np.empty_like(vec2)

    ref = np.isclose(vec1, vec2, **tolerance)
    if mode == 'CPU':
        ct_vec1 = vec1.ctypes.data_as(bin_f.c_func.argtypes[0])
        ct_vec2 = vec2.ctypes.data_as(bin_f.c_func.argtypes[1])
        ct_res = res.ctypes.data_as(bin_f.c_func.argtypes[2])

        bin_f(ct_vec1, ct_vec2, ct_res, len(res))
    else:
        bin_f.cuda_wrap_call(vec1, vec2, res, np.int32(DIM_X))

    assert np.array_equal(res, ref)


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('rtol,atol', [[0, 0], [None, None], [None, 100], [2, None]])
@pytest.mark.parametrize('var1,var2',
                         [[1, 1], [1, 100], [1, 2], [-4, 5], [0, -100], [-1, -2],
                          [[1, 1, 1, -4, 0, -1], [1, 100, 2, 5, -100, -2]]
                         ])
def test_helper_all_close(mode, var1, var2, atol, rtol):

    tolerance = {}
    if rtol is not None:
        tolerance['rtol'] = rtol
    if atol is not None:
        tolerance['atol'] = atol

    vec1 = np.atleast_1d(var1)
    vec2 = np.atleast_1d(var2)
    assert len(vec1) == len(vec2)

    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        arr_ptr_ty = ir.ArrayType(ir.DoubleType(), len(vec1)).as_pointer()
        func_ty = ir.FunctionType(ir.VoidType(), [arr_ptr_ty, arr_ptr_ty, ir.IntType(32).as_pointer()])

        custom_name = ctx.get_unique_name("all_close")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        in1, in2, out = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        all_close = pnlvm.helpers.all_close(ctx, builder, in1, in2, **tolerance)
        res = builder.select(all_close, out.type.pointee(1), out.type.pointee(0))
        builder.store(res, out)
        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)
    vec1 = np.atleast_1d(np.asarray(var1, dtype=bin_f.np_arg_dtypes[0].base))
    vec2 = np.atleast_1d(np.asarray(var2, dtype=bin_f.np_arg_dtypes[1].base))
    assert len(vec1) == len(vec2)
    res = bin_f.np_buffer_for_arg(2)

    ref = np.allclose(vec1, vec2, **tolerance)

    if mode == 'CPU':
        bin_f(vec1, vec2, res)
    else:
        bin_f.cuda_wrap_call(vec1, vec2, res)

    assert np.array_equal(res, ref)

@pytest.mark.llvm
@pytest.mark.parametrize("ir_argtype,format_spec,values_to_check", [
    pytest.param(pnlvm.ir.IntType(32), "%u", range(0, 20), id="i32"),
    pytest.param(pnlvm.ir.IntType(64), "%lld", [int(-4E10), int(-3E10), int(-2E10)], id="i64"),
    pytest.param(pnlvm.ir.DoubleType(), "%lf", [x *.5 for x in range(0, 5)], id="double"),
    pytest.param(pnlvm.ir.FloatType(), "%lf", [x *.5 for x in range(0, 5)], id="float"),
    ])
def test_helper_printf(capfd, ir_argtype, format_spec, values_to_check):

    format_str = f"Hello {(format_spec + ' ') * len(values_to_check)}\n"

    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        func_ty = ir.FunctionType(ir.VoidType(), [])
        ir_values_to_check = [ir_argtype(i) for i in values_to_check]
        custom_name = ctx.get_unique_name("test_printf")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        pnlvm.helpers.printf(ctx, builder, format_str, *ir_values_to_check, tags={"always"})
        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)
    bin_f()

    # Printf is buffered in libc.
    libc = ctypes.util.find_library("msvcrt" if sys.platform == "win32" else "c")
    libc = ctypes.CDLL(libc)
    libc.fflush(0)

    # Convert format specifier to Python compatible
    python_format_spec = {"%lld":"%ld"}.get(format_spec, format_spec)

    # The string below omits the newline character used above and the check below
    # uses 'startswith' to avoid issues with different newline encoding across OSes.
    python_format_str = f"Hello {(python_format_spec + ' ') * len(values_to_check)}"
    assert capfd.readouterr().out.startswith(python_format_str % tuple(values_to_check))

class TestHelperTypegetters:
    FLOAT_TYPE = pnlvm.ir.FloatType()
    FLOAT_PTR_TYPE = pnlvm.ir.PointerType(FLOAT_TYPE)
    DOUBLE_TYPE = pnlvm.ir.DoubleType()
    DOUBLE_PTR_TYPE = pnlvm.ir.PointerType(DOUBLE_TYPE)
    DOUBLE_VECTOR_TYPE = pnlvm.ir.ArrayType(DOUBLE_TYPE, 1)
    DOUBLE_VECTOR_PTR_TYPE = pnlvm.ir.PointerType(DOUBLE_VECTOR_TYPE)
    DOUBLE_MATRIX_TYPE = pnlvm.ir.ArrayType(pnlvm.ir.ArrayType(DOUBLE_TYPE, 1), 1)
    DOUBLE_MATRIX_PTR_TYPE = pnlvm.ir.PointerType(DOUBLE_MATRIX_TYPE)
    INT_TYPE = pnlvm.ir.IntType(32)
    INT_PTR_TYPE = pnlvm.ir.PointerType(INT_TYPE)
    BOOL_TYPE = pnlvm.ir.IntType(1)
    BOOL_PTR_TYPE = pnlvm.ir.PointerType(BOOL_TYPE)

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 0),
        (FLOAT_PTR_TYPE, 1),
        (DOUBLE_TYPE, 0),
        (DOUBLE_PTR_TYPE, 1),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 1),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 1),
        (INT_TYPE, 0),
        (INT_PTR_TYPE, 1),
        (BOOL_TYPE, 0),
        (BOOL_PTR_TYPE, 1),
    ], ids=str)
    def test_helper_is_pointer(self, ir_type, expected):
        assert pnlvm.helpers.is_pointer(ir_type) == expected
        assert pnlvm.helpers.is_pointer(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 1),
        (FLOAT_PTR_TYPE, 1),
        (DOUBLE_TYPE, 1),
        (DOUBLE_PTR_TYPE, 1),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 0),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 0),
        (INT_TYPE, 1),
        (INT_PTR_TYPE, 1),
        (BOOL_TYPE, 1),
        (BOOL_PTR_TYPE, 1),
    ], ids=str)
    def test_helper_is_scalar(self, ir_type, expected):
        assert pnlvm.helpers.is_scalar(ir_type) == expected
        assert pnlvm.helpers.is_scalar(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 1),
        (FLOAT_PTR_TYPE, 1),
        (DOUBLE_TYPE, 1),
        (DOUBLE_PTR_TYPE, 1),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 0),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 0),
        (INT_TYPE, 0),
        (INT_PTR_TYPE, 0),
        (BOOL_TYPE, 0),
        (BOOL_PTR_TYPE, 0),
    ], ids=str)
    def test_helper_is_floating_point(self, ir_type, expected):
        assert pnlvm.helpers.is_floating_point(ir_type) == expected
        assert pnlvm.helpers.is_floating_point(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 0),
        (FLOAT_PTR_TYPE, 0),
        (DOUBLE_TYPE, 0),
        (DOUBLE_PTR_TYPE, 0),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 0),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 0),
        (INT_TYPE, 1),
        (INT_PTR_TYPE, 1),
        (BOOL_TYPE, 1),
        (BOOL_PTR_TYPE, 1),
    ], ids=str)
    def test_helper_is_integer(self, ir_type, expected):
        assert pnlvm.helpers.is_integer(ir_type) == expected
        assert pnlvm.helpers.is_integer(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 0),
        (FLOAT_PTR_TYPE, 0),
        (DOUBLE_TYPE, 0),
        (DOUBLE_PTR_TYPE, 0),
        (DOUBLE_VECTOR_TYPE, 1),
        (DOUBLE_VECTOR_PTR_TYPE, 1),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 0),
        (INT_TYPE, 0),
        (INT_PTR_TYPE, 0),
        (BOOL_TYPE, 0),
        (BOOL_PTR_TYPE, 0),
    ], ids=str)
    def test_helper_is_vector(self, ir_type, expected):
        assert pnlvm.helpers.is_vector(ir_type) == expected
        assert pnlvm.helpers.is_vector(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 0),
        (FLOAT_PTR_TYPE, 0),
        (DOUBLE_TYPE, 0),
        (DOUBLE_PTR_TYPE, 0),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 0),
        (DOUBLE_MATRIX_TYPE, 1),
        (DOUBLE_MATRIX_PTR_TYPE, 1),
        (INT_TYPE, 0),
        (INT_PTR_TYPE, 0),
        (BOOL_TYPE, 0),
        (BOOL_PTR_TYPE, 0),
    ], ids=str)
    def test_helper_is_2d_matrix(self, ir_type, expected):
        assert pnlvm.helpers.is_2d_matrix(ir_type) == expected
        assert pnlvm.helpers.is_2d_matrix(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (FLOAT_TYPE, 0),
        (FLOAT_PTR_TYPE, 0),
        (DOUBLE_TYPE, 0),
        (DOUBLE_PTR_TYPE, 0),
        (DOUBLE_VECTOR_TYPE, 0),
        (DOUBLE_VECTOR_PTR_TYPE, 0),
        (DOUBLE_MATRIX_TYPE, 0),
        (DOUBLE_MATRIX_PTR_TYPE, 0),
        (INT_TYPE, 0),
        (INT_PTR_TYPE, 0),
        (BOOL_TYPE, 1),
        (BOOL_PTR_TYPE, 1),
    ], ids=str)
    def test_helper_is_boolean(self, ir_type, expected):
        assert pnlvm.helpers.is_boolean(ir_type) == expected
        assert pnlvm.helpers.is_boolean(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,expected', [
        (DOUBLE_VECTOR_TYPE, [1]),
        (DOUBLE_VECTOR_PTR_TYPE, [1]),
        (DOUBLE_MATRIX_TYPE, [1, 1]),
        (DOUBLE_MATRIX_PTR_TYPE, [1, 1]),
    ], ids=str)
    def test_helper_get_array_shape(self, ir_type, expected):
        assert pnlvm.helpers.get_array_shape(ir_type(None)) == expected

    @pytest.mark.llvm
    @pytest.mark.parametrize('ir_type,shape', [
        (DOUBLE_VECTOR_TYPE, (1,)),
        (DOUBLE_MATRIX_TYPE, (1,1)),
    ], ids=str)
    def test_helper_array_from_shape(self, ir_type, shape):
        assert ir_type == pnlvm.helpers.array_from_shape(shape, self.DOUBLE_TYPE)


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('op,var,expected', [
    (pnlvm.helpers.tanh, 1.0, 0.7615941559557649),
    (pnlvm.helpers.exp, 1.0, 2.718281828459045),
    (pnlvm.helpers.coth, 1.0, 1.3130352854993313),
    (pnlvm.helpers.csch, 1.0, 0.8509181282393215),
    (pnlvm.helpers.log, 1.0, 0.0),
    (pnlvm.helpers.log1p, 1.0, 0.6931471805599453),
])
@pytest.mark.parametrize('fp_type', [pnlvm.ir.DoubleType(), pnlvm.ir.FloatType()], ids=str)
def test_helper_numerical(mode, op, var, expected, fp_type):
    with pnlvm.LLVMBuilderContext(fp_type) as ctx:
        func_ty = ir.FunctionType(ir.VoidType(), [ctx.float_ty.as_pointer()])

        custom_name = ctx.get_unique_name("numerical")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        in_out = function.args[0]
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        variable = builder.load(in_out)
        result = op(ctx, builder, variable)
        builder.store(result, in_out)

        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)

    res = np.asarray(var, dtype=bin_f.np_arg_dtypes[0])

    if mode == 'CPU':
        bin_f(res)
    else:
        bin_f.cuda_wrap_call(res)

    np.testing.assert_allclose(res, expected)


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('var1,var2,expected', [
    (np.array([1.,2.,3.]), np.array([1.,2.,3.]), np.array([2.,4.,6.])),
    (np.array([1.,2.,3.]), np.array([0.,1.,2.]), np.array([1.,3.,5.])),
    (np.array([[1.,2.,3.], [4.,5.,6.], [7.,8.,9.]]),
     np.array([[10.,11.,12.], [13.,14.,15.], [16.,17.,18.]]),
     np.array([[11.,13.,15.], [17.,19.,21.], [23.,25.,27.]])),
])
def test_helper_recursive_iterate_arrays(mode, var1, var2, expected):
    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        arr_ptr_ty = ctx.convert_python_struct_to_llvm_ir(var1).as_pointer()

        func_ty = ir.FunctionType(ir.VoidType(), [arr_ptr_ty, arr_ptr_ty, arr_ptr_ty])

        custom_name = ctx.get_unique_name("elementwise_op")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        u, v, out = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        with pnlvm.helpers.recursive_iterate_arrays(ctx, builder, u, v, out) as (loop_builder, a_ptr, b_ptr, o_ptr):
            a = loop_builder.load(a_ptr)
            b = loop_builder.load(b_ptr)
            loop_builder.store(loop_builder.fadd(a, b), o_ptr)

        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)

    vec1 = np.asarray(var1, dtype=bin_f.np_arg_dtypes[0].base)
    vec2 = np.asarray(var2, dtype=bin_f.np_arg_dtypes[1].base)
    res = bin_f.np_buffer_for_arg(2)

    if mode == 'CPU':
        bin_f(vec1, vec2, res)
    else:
        bin_f.cuda_wrap_call(vec1, vec2, res)

    assert np.array_equal(res, expected)


_fp_types = [ir.DoubleType, ir.FloatType, ir.HalfType]


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('t1', _fp_types)
@pytest.mark.parametrize('t2', _fp_types)
@pytest.mark.parametrize('val', [1.0, '-Inf', 'Inf', 'NaN', 16777216, 16777217, -1.0])
def test_helper_convert_fp_type(t1, t2, mode, val):
    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        func_ty = ir.FunctionType(ir.VoidType(), [t1().as_pointer(), t2().as_pointer()])
        custom_name = ctx.get_unique_name("fp_convert")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        x, y = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        x_val = builder.load(x)
        conv_x = pnlvm.helpers.convert_type(builder, x_val, y.type.pointee)
        builder.store(conv_x, y)
        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)

    # Get the argument numpy dtype
    np_dt1, np_dt2 = (np.dtype(bin_f.np_arg_dtypes[i]) for i in (0, 1))

    # instantiate value, result and reference
    x = np.asarray(val, dtype=np_dt1)
    y = bin_f.np_buffer_for_arg(1)
    ref = x.astype(np_dt2)

    if mode == 'CPU':
        bin_f(x, y)
    else:
        bin_f.cuda_wrap_call(x, y)

    np.testing.assert_allclose(y, ref, equal_nan=True)


_int_types = [ir.IntType(64), ir.IntType(32), ir.IntType(16), ir.IntType(8)]


@pytest.mark.llvm
@pytest.mark.parametrize('mode', ['CPU', pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('t1', _int_types, ids=str)
@pytest.mark.parametrize('t2', _int_types, ids=str)
@pytest.mark.parametrize('val', [0, 1, -1, 127, -128, 255, -32768, 32767, 65535, np.iinfo(np.int32).min, np.iinfo(np.int32).max])
def test_helper_convert_int_type(t1, t2, mode, val):
    with pnlvm.LLVMBuilderContext.get_current() as ctx:
        func_ty = ir.FunctionType(ir.VoidType(), [t1.as_pointer(), t2.as_pointer()])
        custom_name = ctx.get_unique_name("int_convert")
        function = ir.Function(ctx.module, func_ty, name=custom_name)
        x, y = function.args
        block = function.append_basic_block(name="entry")
        builder = ir.IRBuilder(block)

        x_val = builder.load(x)
        conv_x = pnlvm.helpers.convert_type(builder, x_val, y.type.pointee)
        builder.store(conv_x, y)
        builder.ret_void()

    bin_f = pnlvm.LLVMBinaryFunction.get(custom_name)

    # Get the argument numpy dtype
    np_dt1, np_dt2 = (np.dtype(bin_f.np_arg_dtypes[i]) for i in (0, 1))

    # instantiate value, result and reference
    x = np.asarray(val).astype(np_dt1)
    y = bin_f.np_buffer_for_arg(1)
    ref = x.astype(np_dt2)

    if mode == 'CPU':
        bin_f(x, y)
    else:
        bin_f.cuda_wrap_call(x, y)

    np.testing.assert_array_equal(y, ref)
